/******************************************************************************/
// modul: TAsciiData.cpp   version 1.0b (english)                             //
//                                                                            //
//	  multi-threaded "load ASCII data-matrix from file" class by lore@newty.de //
//   see header-file for further information and disclaimer.                  //
/******************************************************************************/



// constants
#define MODULE_VERSION 	"TAsciiData v2.15b mt"     // module name
#define COLUMN_WIDTH 	11							 		// minimum column-width when saving data-matrix to file


// macros
#define CATCH_INT		catch(int errNo)  { TErrText err; switch(errNo) {
#define THROW_ERR 	} throw err;  }



// includes
#include <stdlib.h>					// due to: 	randomize()
#include <time.h>						// 			randomize()

#include <stdio.h>					// 			sprintf() ...
#include <iomanip.h>					// 			setw()
#include <math.h>						// 			sqrt()
#include <owl\gauge.h>				// 			TGauge
#include <process.h>    			// 			beginthread

#include "asciidata.h"
#include "fileutil.h"




/***********************************************************************************************************/
// constructor: parameters are optional. load/save and calc run multi-threaded if parameters are supplied.
TAsciiData::TAsciiData(TWindow* _pParent /* =NULL */, TFont* _pFont /* =NULL */, TResId _resId/* =0 */)
{
	// 1. copy parameters
   pParent 			= _pParent;
   pFont				= _pFont;
   resId				= _resId;


   // 2. initialize vivs
   ref 				= 1;
	_nDS 				= 0;
   _dim 				= 0;
   vir_nDS			= 0;
   vir_dim			= 0;
   _depth 			= 0;

   data 				= NULL;
	series			= NULL;
   order				= NULL;


	// 3. initialize pointers for statistical values
   mean				= NULL;
   min				= NULL;
   max				= NULL;
   range				= NULL;
   invRange			= NULL;
   dev				= NULL;
   maxEqualsMin	= NULL;



   // 4. initialize flags
   _IsSorted 				= false; 	// true: rows are sorted
   _loaded					= false;		// 		load() has been called
   _IsLocked 				= false;    // 		locked, i.g. calls of sort() etc. cause an exception
   _consistentStatValues= false;    // 		statistical values are calculated and consistent


   // 5. misc
   allocated_nDS 	= 0;
}



/***********************************************************************************************************/
// wrapper to destruct an instance
void TAsciiData::release() const
{
	ref--;                        	// decrement reference counter

   if(ref == 0)   						// if none reference exists anymore ...
   	delete this;   					// ... call private destructor
}



/***********************************************************************************************************/
// private destructor
TAsciiData::~TAsciiData()
{
   // release statistical values
  	delete[] mean;
  	delete[] min;
  	delete[] max;
   delete[] range;
   delete[] invRange;
   delete[] dev;
   delete[] maxEqualsMin;


	// release data-matrix
   // hint: test of 'data' not necessary cause otherwise <allocated_nDS> would be 0. BUT test of '!series'
   //	required cause in this case the memory is already released!
	if(data && !series)
		for(int i=0;i<allocated_nDS;i++)
   		delete[] data[i];

	delete[] data;
   delete[] order;
   delete[] series;
}



/***********************************************************************************************************/
// allocate memory for data-matrix
void TAsciiData::allocateMemory()
{
	allocated_nDS = _nDS; 			// remember #rows for destruction


	data = new float*[_nDS];   	// allocate memory
   for(int i=0;i<_nDS;i++)
   	data[i] = new float[_dim];


   vir_nDS  = _nDS;					// no time-series yet: the virtual rows/column -number equals the real one
   vir_dim	= _dim;
}



/***********************************************************************************************************/
// display the objects state during load/save/calc-operations using the thread-control dialog.
// annotation: this function can be called directly or as a callback-function by the thread-control dialog
void actState(void* _pMyself)
{
	// explicitly cast the void-pointer to the correct type
	TAsciiData* pMyself = (TAsciiData*) _pMyself;


   // define shortcuts
   TThreadCtrlDlg* 	pCtrlDlg = pMyself->pCtrlDlg;
   int & 				dim		= pMyself->vir_dim;
   int & 				nDS		= pMyself->vir_nDS;


	// execute the rest only if the thread-control dialog exists
	if(!pCtrlDlg)
   	return;		// now the function has done nothing :-)


   // two strings which are filled with the status texts
	char szState1[256];
   char szState2[256];

   szState1[0] = '\0';				// "clear" strings
   szState2[0] = '\0';

   static bool toggle = false;	// flag used for "animation" (blinking '??' which indicate progress)
   if(toggle)	toggle = false;
   else			toggle = true;


	// I. compose the texts which are only changed when the objects state changes
	if(pMyself->updated == false)
	{
  		sprintf(szState1, "cols = %d     rows = %d", dim, nDS);		// initialize with default text


      // compose the texts according to the state
		switch(pMyself->state)
   	{
      	case TAsciiData::startup     		:	sprintf(szState1, "checking filesystem integrity ..." /*:-)*/); break;
	 	  	case TAsciiData::detectCols 		:	sprintf(szState1, "detecting number of columns");
	  														sprintf(szState2, "cols = ??     rows = ??");
																						   	     											 break;

   	   case TAsciiData::detectRows 		:	sprintf(szState1, "detecting number of rows ...");    		  	 break;

			case TAsciiData::reading			:	sprintf(szState1, "reading ...");										 break;

			case TAsciiData::conv2timeSeries	:	if(toggle)	sprintf(szState1, "converting to time-series ...");
         												else			sprintf(szState1, "converting to time-series");
		                          					pCtrlDlg->setGauge(100);
   																						                              		    break;

	  		case TAsciiData::calcI				:	sprintf(szState1, "calculating mean ...");  							 break;

			case TAsciiData::calcII				:	sprintf(szState1, "calculating deviation ...");						 break;
 			case TAsciiData::saving				:	sprintf(szState1, "saving ...");											 break;
   	}

      // transfer texts to the dialog
 		pCtrlDlg->setText1(szState1);
		pCtrlDlg->setText2(szState2);

      pMyself->updated = true;       		// reset flag
   }


   // II. compose texts which are to be updated every time
   switch(pMyself->state)
   {
   	case TAsciiData::detectRows	:  // this code provides a blinking '??'
      											if(toggle) sprintf(szState2, "cols = %d     rows = ??", 	dim);
		                                 else       sprintf(szState2, "cols = %d     rows =", 		dim);
													pCtrlDlg->setText2(szState2);                                	break;

		case TAsciiData::reading 	:    	pCtrlDlg->setGauge(100*pCtrlDlg->progress/(nDS+1));  					break;
      case TAsciiData::calcI	 	:		pCtrlDlg->setGauge(100*pCtrlDlg->progress/(2*dim));     				break;
		case TAsciiData::calcII		: 		pCtrlDlg->setGauge(100*pCtrlDlg->progress/(2*dim));     				break;
		case TAsciiData::saving		:		pCtrlDlg->setGauge(100*pCtrlDlg->progress/(nDS+1));     				break;
  	}
}	// end-function actState()





/***********************************************************************************************************/
// wrapper to load data-matrix from file: if a parent has been supplied in constructor, ::load() runs in an
//	additional  thread, otherwise it is called in the mainthread
void TAsciiData::load(const char* filename)
{
 // 1. try to open file for reading
	infile.open(filename, ios::in);


 // 2. check permissibility
	try
   {
		if(_IsLocked)	throw 0;
   	if(_loaded)		throw 1;    // load() already has been called
      if(!infile)		throw 2;		// file open fails
   }
	CATCH_INT
		case 0	: 	sprintf(err.szErrText, "TAsciiData::load() called on locked instance! Call unlock() before!"); 	break;
		case 1	: 	sprintf(err.szErrText, "TAsciiData::load() called twice in the lifetime of an instance!"); 		break;
		case 2	: 	sprintf(err.szErrText, "TAsciiData::load(): unable to open file!");	 									break;
	THROW_ERR


  	_loaded 			= true;			// set flag: prevent this function to be called again
   err.szErrText[0] 	= '\0';  	// reset error text


  // 3. if a parent was supplied in constructor -> ... create thread-control dialog and execute ::load() in
  //	  additional thread
   if(pParent && pFont)
   {
		// a) compose dialogs caption
   	char szText[256];
      sprintf(szText, "loading '%s'", filename);

	   // b) instantiate thread-control dialog
   	pCtrlDlg = new TThreadCtrlDlg(pParent, szText,  resId, pFont);

		//	c) supply the callback function for periodical display of the loading-state
	   pCtrlDlg ->setCallbackFunc(::actState, (void*) this, /* call every 0.15 seconds = */ 150);

		// d) start thread
	   _beginthread(::load, 4096, (void*) this);

   	// e) execute thread-control dialog
		pCtrlDlg->Execute();

      // f) delete dialog
      delete pCtrlDlg;
   }
   else   // else if no parent was supplied -> execute ::load() in main-thread
	{
      pCtrlDlg = NULL;
   	::load((void*) this);
   }


  // 4. close file
   infile.close();


  // 5. check for errors, which are coded in 'szErrText', and throw an exception if one occured
   if(err.szErrText[0] != '\0')
   	throw err;


}	// end-function load()




/***********************************************************************************************************/
//	 load data-matrix from file. #columns and #rows are detected self-acting. exceptions are thrown in case
//	 of errors
//
// 	2. detect #columns: inspect first non empty row.
// 	3. detect #rows: count lines in file. empty lines and comments are ignored
// 	5. allocate memory
//		6. if running in a thread: supply stop-flag to the thread-control dialog
// 	7. read data and check endls
//		8. convert to time-series if its a vector (dim==1)
//		9. if runnig in a thread: close thread-control dialog
void load(void* _pMyself)
{
  // 1. foreplay
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL);			// reduce thread priority


   // shortcuts
   TAsciiData* 		 	pMyself 	= (TAsciiData*) _pMyself;							// explicit typecast
   ifstream& 				file 		= pMyself->infile;
   int &						nDS		= pMyself->_nDS;
   int & 					dim		= pMyself->_dim;
	TThreadCtrlDlg* 		pCtrlDlg	= pMyself->pCtrlDlg;


  	pMyself->setState(TAsciiData::detectCols);		// change state to 'detecting #columns'
  	streampos curPos = file.tellg(); 					// remember begin of the file to restore in step 4


   int line	= 1;	// line counter


  // 2. detect #columns: inspect first non-empty line.
   try
   {
		line += skipwsEx(file);  					// remove all whitespaces and(!) comments

   	label:

		read(file, (float) 1.0);      			// read a float value
		dim++;											// increment #columns
		if(skipws(file)==0)                    // remove all whitespaces and loop back if no endl occurs
	      goto label;

   }
   catch (int errNo)
	{
	   // propagate exception
      sprintf(pMyself->err.szErrText, "detecting number of columns - line %d:   %s", line, ::getErrCode(errNo));

      if(pCtrlDlg)													// if running in a thread: close dialog
			pCtrlDlg->shutDown();

		return;															// end thread/function
   }
   pMyself->vir_dim	= dim;
	pMyself->setState(TAsciiData::detectRows);					// change state to 'detecting #rows'



  // 3. count number of lines, i.g. count the '\n' in the file ignoring empty lines or comments
   nDS = countLinesEx(file)+1;


  // 4. restore file position to the begin
   file.seekg(curPos);


  // 5. allocate memory for data-matrix with size 'nDS' x 'dim'
   pMyself->allocateMemory();


  // 6. if running in thread: supply stop-flag to the thread-control dialog
   bool stop=false;
   if(pCtrlDlg)
	   pCtrlDlg->setStopFlag(&stop);




	pMyself->setState(TAsciiData::reading);	// change state to 'reading rows'


  // 7. read rows
   int actRow, j;
   line = 1;

   try
   {
      line += skipwsEx(file);  											// remove all whitespaces and comments

 		for (actRow=0; actRow<nDS;actRow++)  							// read over all (expected) rows
  		{
			// formatted reading: function can be found in 'fileutil.h'.
         // mechanism: 	a) read all values of a row using the functions skipwsNoEndl() and read(). the first
         //						skips all whitespaces throwing and exception if an endl occurs. the latter then
         //						reads the float-value.
         //					b) call skipwsExEndl() which skips all whitespaces and(!) comments throwing an
         //						exception if no 'endl' is entcountred.

			// a) over all columns
 			for(j=0; j<dim;j++)
			{
          	skipwsNoEndl(file);
          	pMyself->data[actRow][j] = read(file, (float) 0);
         }

         j--;	// just4fun - try to find out why :-)

			// b) check the 'endl'
         line += skipwsExEndl(file);

         if(pCtrlDlg)
			{
	      	pCtrlDlg->progress++;
   	      if(stop)                   // check stop-flag -> ...
   	     		throw errUnexpectedEOF; // ... end throwing an exception. (annotation: exception will be ignored)
         }
    	}
   }
   catch (int errNo)
   {
	   nDS 					= actRow; 			// copy the number of loaded rows
      pMyself->vir_nDS 	= nDS;


      // if a "real" error occured: set error-text which is checked at the end of the wrapper load()
      if (errNo != errUnexpectedEOF)
	      sprintf(pMyself->err.szErrText,"line %d  -  row %d  -  column %d  -  %s", line,
         														 actRow+1, j+1, ::getErrCode(errNo));
   }


  // 8. if the data is a vector (dim==1) -> convert to time-series with depth 0
 	pMyself->setState(TAsciiData::conv2timeSeries);
 	if(dim == 1)
   	pMyself->set_time_series(0);


 // 9. if running in a thread: close thread-control dialog
  	if(pCtrlDlg)
		pCtrlDlg->shutDown();

} // end-function load()




/***********************************************************************************************************/
// wrapper to save data-matrix to file: if a parent has been supplied in constructor, ::save() runs in an
//	additional thread, otherwise it is called in the mainthread
void TAsciiData::save(const char* filename) const
{
  // 1. try to open file for writing
	outfile.open(filename, ios::out);
   if(!outfile)
   {
    	strcpy(err.szErrText, "error: unable to open file");
    	throw err;     // if file open fails -> throw exception
   }

  // 2. initialize for additional thread
   err.szErrText[0] = '\0';


  // 3. if a parent was supplied in constructor -> ... create thread-control dialog and execute ::save()
  //	  in additional thread
   if(pParent && pFont)
   {
		// a) compose dialogs caption
   	char szText[256];
      sprintf(szText, "saving '%s'", filename);

	   // b) instantiate thread-control dialog
   	pCtrlDlg = new TThreadCtrlDlg(pParent, szText,  resId, pFont);

		//	c) supply the callback function for periodic display of the saving-state
	   pCtrlDlg ->setCallbackFunc(::actState, (void*) this, /* time = */ 150);

		// d) start thread
	   _beginthread(::save, 4096, (void*) this);

   	// e) execute thread-control dialog
		pCtrlDlg->Execute();

      // f) delete dialog
      delete pCtrlDlg;
   }
   else     // else if no parent was supplied -> execute ::load() in main-thread
	{
      pCtrlDlg = NULL;
   	::save((void*) this);
   }

   // 4. close file
   outfile.close();


  	// 5. check for errors, which are coded in 'szErrText', and throw an exception if one occured
	if(err.szErrText[0] != '\0')
   	throw err;

}	// end-function save()



/***********************************************************************************************************/
// save data-matrix to file
void save(void* _pMyself)
{
  // 1. foreplay
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL);		// reduce thread-priority

   // shortcuts
   TAsciiData*   	pMyself 	= (TAsciiData*) _pMyself; 								//	explicit typecast
	float**			data		= pMyself->data;
   ofstream& 		file 		= pMyself->outfile;
   int 				vir_nDS	= pMyself->vir_nDS;
   int  				vir_dim	= pMyself->vir_dim;
   float*& 	mean 				= pMyself->mean;
   float*& 	max 				= pMyself->max;
   float*& 	min 				= pMyself->min;
   float*& 	dev 			   = pMyself->dev;
	TThreadCtrlDlg* pCtrlDlg	= pMyself->pCtrlDlg;


   // 2. write header
  	file << "* data written by " << MODULE_VERSION << endl;
   file << "* #rows = " << vir_nDS << "  #columns = " << vir_dim << "  (memory was allocated for ";
   file << pMyself->allocated_nDS << " rows)" << endl;


	pMyself->setState(TAsciiData::saving);		// change state to 'saving to file'


  // 3. if running in thread: supply stop-flag to the thread-control dialog
   bool stop = false;
	if(pCtrlDlg)
	   pCtrlDlg->setStopFlag(&stop);


  // 4. if the statistical, column-wise calculated, values exist: write them
   if(mean && pMyself->_consistentStatValues)
   {
  	   file << endl << "* minimal values" << endl << "* ";	  // minima
		for(int j=0;j<vir_dim;j++)	 file << setw(COLUMN_WIDTH) << min[j] << " ";
	   file << endl;

      file << endl << "* maximal values" << endl << "* ";	  // maxima
		for(int j=0;j<vir_dim;j++)	file << setw(COLUMN_WIDTH) << max[j] << " ";
     	file << endl;

     	file << endl << "* mean values" << endl << "* ";		  // means
		for(int j=0;j<vir_dim;j++)	file << setw(COLUMN_WIDTH) << mean[j] << " ";
      file << endl;

     	file << endl << "* standard deviation" << endl << "* "; // deviation
		for(int j=0;j<vir_dim;j++)	file << setw(COLUMN_WIDTH) << dev[j] << " ";
     	file << endl << endl;
   }


  // 5. write all rows
  	for(int i=0;i<vir_nDS;i++)
	{
      if(pCtrlDlg)
      	pCtrlDlg->progress++;

      // write all columns
  		for(int j=0;j<vir_dim;j++)
 			file << setw(COLUMN_WIDTH) << data[i+pMyself->_depth][j] << " ";
      file << endl;


      // check the stop-flag and if true -> end loop
      if(stop)
		{
       	file << endl << "error: userbreak!!  process terminated!!" << endl;
      	break;
      }
  	}


 // 6. if running in a thread: close thread-control dialog
   if(pCtrlDlg)
		pCtrlDlg->shutDown();

} // end-function save()




/***********************************************************************************************************/
// wrapper for the column-wise calculation of statistical values. if a parent has been supplied in
// constructor, ::calc() runs in an additional thread, otherwise it is called in the mainthread
void TAsciiData::calc() const
{
	// 1. check #rows cause the calculation of the deviation needs at least two values/rows
	if(_nDS < 2)
   {
      strcpy(err.szErrText, "not enough datatuples to calculate standard deviation!");
   	throw err;
   }

  	_consistentStatValues = false;	// reset flag: the values will not be consistent during calculation

  // 2. if a parent was supplied -> create thread-control dialog and execute ::calc() in additional thread
   if(pParent && pFont)
   {
	   // a) instantiate thread-control dialog
   	pCtrlDlg = new TThreadCtrlDlg(pParent, "calculating statistical values",  resId, pFont);


		//	b) supply the callback function for periodic display of the calculation-state
	   pCtrlDlg ->setCallbackFunc(::actState, (void*) this, /* time = */ 150);

		// c) start thread
	   _beginthread(::calc, 4096, (void*) this);

   	// d) execute thread-control dialog
		pCtrlDlg->Execute();

      // e) delete dialog
      delete pCtrlDlg;
   }
   else   // else if no parent was supplied -> execute ::calc() in main-thread
	{
      pCtrlDlg = NULL;
   	::calc((void*) this);
   }

	_consistentStatValues = true;
  	// annotation: there is no error-checking cause no errors are designated ;-)
}



/***********************************************************************************************************/
// column-wise calculate statistical values: mean, max, min, range and deviation
void calc(void* _pMyself)
{
 // 1. foreplay
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL);		// reduce thread priority


   // shortcuts
	TAsciiData* pMyself 			= (TAsciiData*) _pMyself; 							// explicit typecast
   float*& 		mean 				= pMyself->mean;
   float*& 		max 				= pMyself->max;
   float*& 		min 				= pMyself->min;
   float*& 		dev 				= pMyself->dev;
   float*& 		range 			= pMyself->range;
   float*& 		invRange   		= pMyself->invRange;
   bool*& 		maxEqualsMin 	= pMyself->maxEqualsMin;
   int 		   vir_dim 			= pMyself->vir_dim;
   int 			vir_nDS			= pMyself->vir_nDS;
   int 			_depth			= pMyself->_depth;
   float**  	data     		= pMyself->data;

   TThreadCtrlDlg* pCtrlDlg	= pMyself->pCtrlDlg;


   // ... release "old" memory
   delete[] mean;
   delete[] min;
   delete[] max;
   delete[] dev;
   delete[] range;
	delete[] invRange;
   delete[] maxEqualsMin;


  	// allocate "new" one
	mean 				= new float[vir_dim];
   max   			= new float[vir_dim];
	min				= new float[vir_dim];
	dev				= new float[vir_dim];
  	range		 		= new float[vir_dim];
   invRange			= new float[vir_dim];
	maxEqualsMin	= new bool[vir_dim];


	pMyself->setState(TAsciiData::calcI);		// change state to 'calculating mean, max and min'


  // 2. calculate mean, max and min over all columns
   for(int j=0;j<vir_dim;j++)
	{
      if(pCtrlDlg)
			pCtrlDlg->progress++;

		// initialize
		mean	[j]		= 0;
      max	[j]		= data[0+_depth][j];
      min	[j]		= data[0+_depth][j];
      dev[j] =0;										// initialize for 3.


   	for(int i=0;i<vir_nDS;i++)					// over all rows
      {
			mean[j]	+= data[i+_depth][j];
         if (data[i+_depth][j] > max[j]) 	max[j] = data[i+_depth][j];
			// the 'else' is saving a few time but it necessitates(!) the ini above
         else
         	if (data[i+_depth][j] < min[j]) 	min[j] = data[i+_depth][j];
      }
		mean[j]	= mean[j]/vir_nDS;

      // for each column: check if the maximum equals the minimum and set the corresponding flag if true
    	if(max[j]==min[j])		maxEqualsMin[j] = true;
      else              		maxEqualsMin[j] = false;


      // calculate range and inverse of range
   	range[j] 	= max[j]-min[j];
      invRange[j] = 1/range[j];
    }



  // 3. change state to 'calculating deviation'
	pMyself->setState(TAsciiData::calcII);

   for(int j=0;j<vir_dim;j++)			// over all columns
	{
      if(pCtrlDlg)
			pCtrlDlg->progress++;

		for(int i=0;i<vir_nDS;i++)			// over all rows
         dev[j]	+=	(mean[j]-data[i+_depth][j])*(mean[j]-data[i+_depth][j]);
   }
	for(int j=0;j<vir_dim;j++)
   	dev[j] = sqrt(dev[j]/(vir_nDS-1));


  // 4. if running in a thread: close thread-control dialog
	if(pCtrlDlg)
		pCtrlDlg->shutDown();

}	// end-function calc()




/***********************************************************************************************************/
// returns true if the maximum equals the minimum in column 'col'
bool TAsciiData::maxEqual2Min(const int column) const
{
	// check permissibility of 'col'
	if(column < 0 || column > vir_dim)
	{
      strcpy(err.szErrText, "index exceeds matrix dimensions in call to TAsciiData::maxEqual2Min()");
   	throw	err;
   }

	return maxEqualsMin[column];
}


/***********************************************************************************************************/
// randomize order of rows by '_n' times interchanging two rows
void TAsciiData::randomize(const int _n) const
{
  	::randomize(); 				// initialize

	// if no vector with the original order exists: create and initialize it
	if(!order)
	{
	   order = new int[_nDS];
   	for(int i=0;i<_nDS;i++)
   		order[i] = i;
   }


   // randomization loop
  	for(int i=0;i< _n;i++)		// interchange '_n' times
	{
		int d1 = random(vir_nDS)+_depth; 	// select two random rows
		int d2 = random(vir_nDS)+_depth;


      // important hint for time-series functionality: the first '_depth' rows are not changed!

      // interchange rows
		float* tmp = data[d1];
      data[d1] = data[d2];
	   data[d2] = tmp;

		// interchange the indices in vector with original order analogue
		int tmp2 = order[d1];
      order[d1] = order[d2];
	   order[d2] = tmp2;
   }

   _IsSorted = false;		// reset flag
}



/***********************************************************************************************************/
// restore original order of rows
void TAsciiData::reorder() const
{
	// if vector with original order exists -> ... restore
 	if(order)
   {
   	float** tmp = new float*[_nDS];	// temp vector to which ...

		for(int i=0;i<_nDS;i++)				// ... the pointers of the rows are written in original order
      	tmp[order[i]] = data[i];

		for(int i=0;i<_nDS;i++)				// back-copy
      	data[i] = tmp[i];


		// release memory
      delete[] tmp;
      delete[] order;
      order = NULL;

	   _IsSorted = false;					// reset flag
   }
}



/***********************************************************************************************************/

	///////////////////////
	typedef struct
   {
   	float value;
		int	index;

	} TIndexAndValue;


	/////////////////////////////////////////////////////////////////////////////////
	// comparision function for the sort algorithm. (callback function)
	int cmp(const void* _a, const void* _b)
	{
		// never forget to change this cast if you change the type of the sorted items
   	const TIndexAndValue* a = (const TIndexAndValue*) _a;
	   const TIndexAndValue* b = (const TIndexAndValue*) _b;

	   if(a->value > b->value)	return 1;
   	else
   		if(a->value == b->value)
	      	if(a->index > b->index)		// in case of equality: keep their order
   	      	return 1;
      	   else
	         	return -1;
   	   else
      		return -1;
	}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// sort the rows according to the values of column 'col'
void TAsciiData::sort(const int col) const
{
	// 1. check permissibility
	try
   {
	   if(col < 0)				throw 1;
	   if(col >= vir_dim)	throw 2;
   }
	CATCH_INT
		case 1	:	sprintf(err.szErrText, "TAsciiData:sort()  -  negative column specified");							break;
		case 2	:	sprintf(err.szErrText, "TAsciiData:sort()  -  column cannot be greater than %d", vir_dim-1);	break;
   THROW_ERR



   // 2. initialize
	TIndexAndValue* ord = new TIndexAndValue[_nDS];

   for(int i=0;i<_depth;i++)              //	remember: the first depth rows must not(!) be affected
   	ord[i].index = i;

   for(int i=_depth;i<_nDS;i++)
	{
   	ord[i].value = data[i][col];
		ord[i].index = i;
   }


   // 3. sort mapper using qsort()
 	qsort((void*) &ord[_depth], vir_nDS, sizeof(ord[0]), cmp);


   // 4. use mapper
   float** tmp = new float*[_nDS];											// vector to which ...

	for(int i=_depth;i<_nDS;i++)	tmp[i]  = data[ord[i].index];		// ... the values are temporarily copied
	for(int i=_depth;i<_nDS;i++) 	data[i] = tmp[i];						// back-copy

   delete[] tmp;



   // 5. set the values of the vector with original order
	if(!order)
	{
	   order = new int[_nDS];  			// if it doesnt exist: create direct
   	for(int i=0;i<_nDS;i++)
   		order[i] = ord[i].index;
   }
   else	// else: interchange entries analoguos to 4. using the mapper 'ord'
   {
   	int* tmp = new int[_nDS];

      for(int i=_depth;i<_nDS;i++)	tmp[i] 	= order[ord[i].index];
		for(int i=_depth;i<_nDS;i++)	order[i] = tmp[i];					// back-copy

    	delete[] tmp;
   }
   delete[] ord;


   _IsSorted = true;		// set flag
}	// end-function sort()



/***********************************************************************************************************/
// interpret vector (dim==1) as time-series   -  note: see description at the top of the header-file
void TAsciiData::set_time_series(const int __depth)
{
 // 1. check premissibility
	try
   {
		if(_IsLocked)							throw 0;
   	if(_dim != 1)							throw 1;
	   if(__depth < 0)     					throw 2;
      if(__depth > getMaxDepth())     	throw 3;
      if(order)						     	throw 4;
   }
	CATCH_INT
		case 0	: 	sprintf(err.szErrText, "TAsciiData::set_time_series() called on locked instance!! Call unlock() before!!"); break;
     	case 1	:	sprintf(err.szErrText, "I can only convert one dimensional data to time-series!");	 					  		  break;
		case 2	:	sprintf(err.szErrText, "Time-series cannot have negative depths!"); 										  			  break;
  		case 3 	: 	sprintf(err.szErrText, "Not enough data-tuples to convert to time-series with depth %d!", __depth);			  break;
		case 4	:	sprintf(err.szErrText, "Call TAsciiData::reorder() before converting to time-series!"); 						  break;
   THROW_ERR


   _consistentStatValues = false; 	// reset flag:  #columns changed, thus the values are no longer consistent


	_depth 	= __depth;					// set depth
   vir_nDS 	= _nDS - _depth;  		// set the virtual number of rows
   vir_dim  = _depth+1;					// ... and columns


 // 3. if the data already has been converted to a series -> ready = return
   if(series)
   	return;


 // 4. else convert
   series = new float[_nDS];			// allocate memory for 'series'


	// copy data in inverse order to the series-vector
   for(int i=0;i<_nDS;i++)
	{
   	series[_nDS-i-1] 		= data[i][0];    			// copy data
    	delete[] 				data[i];						// ... release ...
		data[i] 					= &series[_nDS-i-1];  	// assign new adress
   }
}



/***********************************************************************************************************/
// set state of the object and call actState() to ensure immediate update of the thread-control dialog
void TAsciiData::setState(TStateEnum _state)
{
	state 	= _state;			// set state
   updated  = false;          // set flag that the state has changed
	actState((void*) this);		// use actState() to update the display
}


/***********************************************************************************************************/
// column-wise standardization of column 'col' or of all columns (default).
// the data is standardized to the intervall [0..1] if 'zero2One' is true, otherwise it is standardized in
// such a way that the mean equals zero and 'mean+deviation' equals one.
void TAsciiData::standardize(const bool zero2One, const int col /* =-1 */)
{
 // 1. check permissilibility
	try
   {
	   if(_IsLocked)   							throw 0;
		if(series && _depth != 0)     		throw 1;
   	if(col < 	0 && col != -1)  			throw 2; // '-1' signals, that all columns shall be standardized
	   if(col >= _dim)    						throw 3;
      if(!mean && _consistentStatValues)	throw 4;
   }
	CATCH_INT
		case 0	: 	sprintf(err.szErrText, "TAsciiData::standardize() called on locked instance!! Call unlock() before!!"); 	break;
     	case 1	:	sprintf(err.szErrText, "Its only possible to norm a time-series with depth = 0!");								break;
		case 2	:	sprintf(err.szErrText, "TAsciiData::standardize() : Column cannot be negative!!"); 								break;
  		case 3 	: 	sprintf(err.szErrText, "TAsciiData::standardize() : Column cannot be greater than %d!", _dim);				break;
  		case 4 	: 	sprintf(err.szErrText, "TAsciiData::standardize() : You have to call calc() before!");							break;
	THROW_ERR


   _consistentStatValues = false;       // reset flag: the statistic values are no longer consistent


 // 2. set subtraktor and denominator for the choosed method
	const float* sub;		// subtraktor
   const float* den;		// denominator

   //  set 'sub' and 'den' according to method
   if(zero2One)
   {
   	sub = getMinVec();
      den = getRangeVec();
   }
   else
   {
   	sub = getMeanVec();
      den = getDevVec();
   }

 // 3.standardize
  	if(col != -1)
      for(int i=0;i<_nDS;i++)												// specific column
         data[i][col] = (data[i][col]-sub[col])/den[col];
 	else
   	for(int i=0;i<_nDS;i++)                                  // all columns
      	for(int j=0;j<_dim;j++)
	         data[i][j] = (data[i][j]-sub[j])/den[j];


}	// end-function standardize()





