/******************************************************************************/
// modul: fileutil.cpp  version 1.0b (english)                                //
//                                                                            //
//   utility functions to read values/strings from a stream. by lore@newty.de //
//   see header-file for further information and disclaimer.                  //
//                                                                            //
/******************************************************************************/

#include <ctype.h> 					//	due to: 	isspace()
#include <limits.h>					//				INT_MAX
#include <stdio.h>					//				sprintf()


#include "fileutil.h"

#define LEN_ILLEGAL_SEQ 10



/***********************************************************************************************************/
// global variables
static char _errCode		[256];   // error-text
static char _illSeq		[256];	// illegal sequence
static char	_illChar	='\0';		// illegal character

char& illChar() { return _illChar; }


/***********************************************************************************************************/
// function to get a description of the last error.
const char* getErrCode(const int errNo)
{
   switch(errNo)
   {
		case errNoWhitespace			:
      	switch(_illChar)
         {
         	case '.'		: sprintf(_errCode,"Point directly follows preceeding number! \nMaybe youve read a 'float' as an 'int'?");	break;
				case '+'		:
				case '-'		: sprintf(_errCode,"'+' or '-' directly follows preceeding number!"); 	break;
            default		: sprintf(_errCode,"Encountered illegal character '%c'!", _illChar);
         }  break;

    	case errUnexpectedEndl 		: 	sprintf(_errCode,"Unexpected end of line!");		break;
		case errUnexpectedEOF		: 	sprintf(_errCode,"Unexpected end of file!");		break;
      case errCharacterIllegal	:
         if (ispunct(_illChar))
            switch(_illChar)
            {
            	case '+' :
               case '-' : 	sprintf(_errCode,"Displaced character '%c' - maybe a blank to much?", _illChar);	break;
               default 	:  sprintf(_errCode,"Encountered illegal character '%c'!", _illChar); 					break;
            }
         else
      		sprintf(_errCode,"Encountered illegal character '%c'!", _illChar);
         break;

		case errNoEndlFound			:
      	_illSeq[LEN_ILLEGAL_SEQ]='\0';	  // terminate the string
      	sprintf(_errCode, "Line should be ended but I found '%s'!", _illSeq);       								break;
      default					  		:
     		sprintf(_errCode,"Unknown reason!");			  																		break;
 	}

   return _errCode;
}



/***********************************************************************************************************/
// utility function: returns true if 'c' is a comment character, e.g. a character starting a comment
inline bool IsComment(const char c) { return (c == '#' || c == '%' || c == '*'); }




/***********************************************************************************************************/
// wrapper which removes all whitespaces from a stream. an exception is thrown if no endl is encountered
int skipwsEndl(ifstream& file)
{
  	int lines =skipws(file);        	// remove all whitespaces

   if (lines==0)
   {
		file.get(_illSeq, 256);
      throw errNoEndlFound; 			// throw exception if no endl occured
   }

   return lines;
}



/***********************************************************************************************************/
// like skipwsEndl() but the function skipwsEx(), which addionally removes comments, is used
int skipwsExEndl(ifstream& file)
{
  	int lines =skipwsEx(file);        	// remove all whitespaces and(!) comments

   if (lines==0)
   {
		file.get(_illSeq, 256);
      throw errNoEndlFound; 			// throw exception if no endl occured
   }

   return lines;
}



/***********************************************************************************************************/
// wrapper which removes all whitespaces from a stream. an exception is thrown if an endl is encountered
void skipwsNoEndl(ifstream& file)
{
  	int lines =skipws(file);        	// remove all whitespaces

   if (lines>0)
      throw errUnexpectedEndl; 			// throw exception if an endl occured
}



/***********************************************************************************************************/
// like skipwsNoEndl() but the function skipwsEx(), which addionally removes comments, is used
void skipwsExNoEndl(ifstream& file)
{
  	int lines =skipwsEx(file);        	// remove all whitespaces and(!) comments

   if (lines>0)
      throw errUnexpectedEndl; 			// throw exception if an endl occured
}



/***********************************************************************************************************/
// functions reads all whitespaces from an ifstream returning the number of encountered endls
int skipws(ifstream& file)
{
	file.clear();							// reset stream state

	int line =0;
   int c;

   // remove whitespaces counting endls
   do
   {
		c=file.get();
		if(!isspace(c))
      	break;

		if (c =='\n')	  				// if its an endl: increment counter
      	line++;

   } while (file.fail()==0);		// proceed while there are ws in the stream


	// finish
	file.clear();					// reset stream state
   file.putback((char) c);		// put the last character back in the stream cause its no ws

   return line;
}



/***********************************************************************************************************/
// functions reads all whitespaces and comments from an ifstream returning the number of encountered endls.
int skipwsEx(ifstream& file)
{
	file.clear();							// reset stream state

	int line =0;
   int c;

	// while there are whitespaces or comments ...
   do
	{
		c =file.get();                         // read char and ...
 		if(IsComment((char) c))								// ... if its a comment ->  ...
      {
		   file.ignore(INT_MAX, '\n');			// ... skip the rest of the line
      	line++;											// increment line-counter
      }
      else
	   	if (c =='\n')								// else if its an endl
   	   	line++;										// increment line-counter
         else
         	if(!isspace(c))                  // else if its neither ws nor comment -> break
            	break;

   }
   while (file.fail()==0);


	// finish
	file.clear();					// reset stream state
   file.putback((char) c);		// put the last character back in the stream


   return line;
}


/***********************************************************************************************************/
// utility-function for template readExpNoEndl() - is called to examine a stream failstate.
void readFailed(ifstream& file)
{
 	file.clear();							// reset stream state

   int c;

   if (EOF == (char) file.peek())  	// peek and compare with EOF (end of file)
     	throw errUnexpectedEOF;


  // else: inspect if the character is a ws. if so it cannot be the one who caused the error. so go back
  // in the stream until you find a none ws. this is the one the user wants to know;-)
  while(isspace(c = file.get()) && file.fail()==0)
      file.seekg(-2,ios::cur);

   _illChar = (char) c;


  	throw errCharacterIllegal;		// throw exception
}



/***********************************************************************************************************/
// function reads a string from an ifstream. the characters '=' and ':' and all whitespaces terminate the
// string. a whitespaces is put back in the stream, '=' and ':' are removed. the delimiter is returned.
// note: remind that a string can be empty. the whitespaces are put back cause it is expected to use the
// functions skipws()/skipwsEx() or their wrappers to check format and actual line.
char readString(ifstream& file, char* string, const int len)
{
	file.clear();							// reset stream state

   int i=0, c;

   do
	{
   	c =file.get();						// read character

      if(file.fail()!=0)         	// if the stream gets into a fail state
      {
      	file.clear();						// just for principle: reset stream state
         throw errUnexpectedEOF;			// throw an exception. note: EOF is the only possible reason
      }

		// if character is whitespace or '=' or ':' -> stop reading, e.g. the terminate string and return
		if(c=='=' || c==':')
      	break;

      if(isspace(c) || EOF == (char) c)
      {
      	file.putback((char) c);
      	break;
      }

      string[i++]= (char) c;			// else: copy character to string and continue

	} while(i<len-1);


   string[i]='\0';

   return (char) c;		// return delimiter
}




/***********************************************************************************************************/
// function counts the number of endls in an ifstream up from the actual position. empty lines and comments
// are ignored. comments must start at the begin of a line.
int countLinesEx(ifstream& file)
{
	file.clear();							// reset stream state

	int line =0;
   int c;

  	skipwsEx(file);						// remove all ws and(!) comments

   do
   {
		c=file.get();						// read character
		if (c =='\n')						// if its an endl -> ...
	   {
   		line++;								// increment line-counter
	      skipwsEx(file);					// now were at the begin of a new line: remove all ws and(!) comments
      }
   }
	while(file.fail()==0);


   // finish
	file.clear();					// reset stream state
  	file.putback((char) c);		// put the last character back in the stream


   return line;
}


